/*
 * event.c -- scheduler for rlc
 *
 * This file is part of
 *
 * rlc -- Reliable Multicast data Distribution Protocol
 * 
 * (C) 1996, 1997, 1998 Luigi Rizzo and Lorenzo Vicisano
 *     (luigi@iet.unipi.it, vicisano@cs.ucl.ac.uk)
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Luigi Rizzo,
 *      Lorenzo Vicisano and other contributors.
 * 4. Neither the name of the Authors nor the names of other contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

/*** 
 ***  

These routines implement an event loop. First, the event list must
be created using rlc_createevlist(). Timeouts can be registered with
rlc_settimer(). They eventually expire, unless removed using rlc_deletetimer().
rlc_deletetimer() removes all the pending timeouts with a given id.
i/o descriptors can be inserted/deleted using rlc_insertchan() and
rlc_deletechan().

One function per timeout, and one function per channel, can be called.

 ***
 ***/

#include "event.h"
#include "utils.h"
#include "error.h"

#undef TEST
#undef ONE_GO		/* define to make rlc_getevent return after each event */
			/* not sure whether  it still works ... */


static timit_t *timit_free_list = NULL;
static int timit_free_count = 0 ;

static void
free_timit(timit_t *p)
{
    p->next = timit_free_list ;
    timit_free_list = p;
    bzero(p, sizeof (timit_t) );
    timit_free_count++ ;
}

static timit_t *
new_timit()
{
    timit_t *p;

    if (timit_free_list) {
	p = timit_free_list ;
	timit_free_list = p->next ;
	timit_free_count-- ;
    } else {
        p = (timit_t *)calloc(1, sizeof(timit_t)) ;
	if (p == NULL) {
	    (void) rlc_internal_set_error_(-1, ALLOC_ERROR, NULL);
	    return NULL;
	}
    }
    return p ;
}

void *
rlc_createevlist_()
{
    int i;
    evlist_t *e;

    if (NULL==(e=(evlist_t *)calloc(1, sizeof(evlist_t)))) {
	rlc_internal_set_error_ (-1, NO_EVLIST, NULL);
	return NULL;
    }

    for (i=0; i<MAX_CHAN; i++) {
	e->io_chan[i]= -1;
    }
    e->io_ch_num = 0;
    e->timevent = NULL;
    FD_ZERO(&(e->in_s));
    FD_ZERO(&(e->out_s));
    FD_ZERO(&(e->err_s));
    timestamp_t(&(e->lasttime));
    e->late = 0;
    e->selected = -1;
    e->def_f = NULL;
    return((void *)e);
}

/*
 *        |<--------------- past ----------->|
 *        |<--------- delta --------->|
 *        |---------------------------#----------->  virtual time
 *        ^                           ^
 *     VT start                  first event
 *
 *        |<- late ->|
 *    ----|----------*-----------------------*---->  actual time
 *                   ^                       ^
 *                lasttime                thistime
 */

/*++++
 Selects the next event to be executed or an IO descriptor ready
 for reading, and executes the associate function. If there are no
 events ready, it generally blocks, unless there are no timers
 pending and blocking = 0, in this case it just polls the IO descriptor
 and returns if none is ready for reading.

 Params:
	vp : event manager reference.
	blocking : if = 0 doesn't blok in the select if
		there are no timers, it just polls.

 Returns:
	The id of the last event found or, if none,
		one of the following values (< 0):
		- EV_DEFTIMEOUT  got the default timeout;
		- EV_GOTSIGNAL  got a signal in select();
		- EV_TERMINATE  no event selected because there
		are no timer pending and the IO descriptor list is
		empty, typically we must terminate the event loop;
		- EV_NO_PENDING  no event selected because there
		are no timer pending and no IO descriptor was ready
		(only if blocking = 0).
*/
int
rlc_getevent_(void *vp, int blocking)
{
#define	e ((evlist_t *)vp)
    struct timeval thistime, interval, *timp;
    ui32 past;
    timit_t *tim;
    int id,
	max_chan,
	i, j, ret;

again:
    NDEB(fprintf(stderr, "# %u usec past\n", past));

    tim = e->timevent;
    id = EV_NO_PENDING ;

    NDEB( { timit_t *ti = e->timevent;
	    fprintf(stderr, "X");
	    for ( ; ti != NULL ; ti = ti->next )
		fprintf(stderr, "--->[%u] (%u)|", ti->delta, ti->id);
	    fprintf(stderr, "\n");
	} );

    while (tim != NULL) {

	id = EV_NO_PENDING ;
	/* see how much time has past since last update */
	timestamp_t(&thistime);
	past = diff_t(thistime, e->lasttime) + e->late ;

	if (tim->delta <= past) {
	    e->late = past - tim->delta;  /* how late was this event ? */
	    if (e->late > 1000) { /* more than 1ms late... */
		NDEB(fprintf(stderr, "--- we are %u usec late with ev. %d\n",
			e->late, tim->id));
	    }
	    e->timevent = tim->next;	/* first adj the event list
				       (which could be acceded) */
	    e->lasttime = thistime ;
	    if (tim->t_f)
		tim->t_f(tim->t_p);	/* execute the action */

	    past -= tim->delta;		/* update virtual time */
#ifdef NODEF /* we don't free the arguments */
	    free(tim->t_p);
#endif
	    id = tim->id;
	    free_timit(tim);
	    tim = e->timevent;
#ifdef ONE_GO
	    return(id);
#endif
	} else
	    /* e->late = past ; */
	    break ;
    }

    /* before sleeping adjust e->late to compensate
       the time spent in the last event */
    NDEB(fprintf(stderr, " Before adj late = %d\n", e->late));
    timestamp_t(&thistime);
    e->late += diff_t(thistime, e->lasttime);
    e->lasttime = thistime ;
    NDEB(fprintf(stderr, " After adj late = %d\n", e->late));

    /* set timeout to next event */
    timp = &interval;
    if (e->timevent != NULL) {
	ui32 next;
	if (e->late > e->timevent->delta)
	    goto again;				/* no need to sleep */
	next = e->timevent->delta - e->late;
	interval.tv_sec = next/1000000;
	interval.tv_usec = next%1000000;
	NDEB(fprintf(stderr, "# timeout in %u usec\n", next));
    } else {
	interval = e->default_timeout ;
	if (blocking && e->default_timeout.tv_sec==0
		&& e->default_timeout.tv_usec==0) timp = NULL;
	NDEB(if(timp) fprintf(stderr, "# def timeout in %lu usec\n", 
		e->default_timeout.tv_usec+1000000*e->default_timeout.tv_sec));
    }

    FD_ZERO(&(e->in_s));
    max_chan =  -1 ;
    for (j=0; j< e->io_ch_num; j++)
	if ( (i = e->io_chan[j]) >=0 ) {
	    if (i > max_chan) max_chan = i ;
	    FD_SET( i, &(e->in_s));
	    NDEB(fprintf(stderr, "selecting with %d\n", i ));
	}
    if (e->io_ch_num || e->timevent != NULL) {
	ret=select(1+max_chan, &(e->in_s), &(e->out_s), &(e->err_s), timp);
	NDEB(fprintf(stderr, "select returning...\n"))
	if (ret<0) {		/* signal... */
	    fprintf(stderr,
		"rlc_getevent SIGNAL during poll, shouldn't happen...\n");
	    id = EV_GOTSIGNAL ;
	} else if (ret==0) {	/* timeout */
	    NDEB(fprintf(stderr,"timeout, ev %x blocking %d\n",
		e->timevent, blocking));
	    if (e->timevent != NULL) {
		if (e->def_f)
		    e->def_f(e->def_p);
		goto again;
	    } else
		id = EV_DEFTIMEOUT ;
	} else {			/* descriptor ready */
	    for (i=0; !(e->io_chan[i]>=0
		&& FD_ISSET(e->io_chan[i], &(e->in_s)))
		&& i<MAX_CHAN; i++);
	    e->selected = e->io_chan[i];
	    if (e->io_f[i])
		e->io_f[i](e->io_p[i]);
	    id = e->io_id[i];
	}
    } else if (id == EV_NO_PENDING)
	id = EV_TERMINATE;

    if (e->def_f)
	e->def_f(e->def_p);
    return(id);
}

int
rlc_poolchan_(void *vp, int *rid, void **rp)
{
#define	e ((evlist_t *)vp)
  int i, j, ret;
  int max_chan = 0;
  struct timeval tim;
    bzero(&tim, sizeof(tim));
    if (e->io_ch_num) {
	for (j=0; j< e->io_ch_num; j++)
	    if ( (i = e->io_chan[j]) >=0 ) {
		if (i > max_chan) max_chan = i ;
		FD_SET( i, &(e->in_s));
		NDEB(fprintf(stderr, "selecting with %d\n", i ));
	    }
	ret=select(1+max_chan, &(e->in_s), NULL, NULL, &tim);
	NDEB(fprintf(stderr, "select returning...\n"))
	if (ret>0) {		/* descriptor ready... */
	    for (i=0; !(e->io_chan[i]>=0
		&& FD_ISSET(e->io_chan[i], &(e->in_s)))
		&& i<MAX_CHAN; i++);
	    e->selected = e->io_chan[i];
	    *rid = e->io_id[i];
	    *rp = e->io_p[i];
	    return (e->selected);
	}
    }
    return (-1);
}


void
rlc_set_default_timeout_(void *vp, unsigned long delta)
{
#define	e ((evlist_t *)vp)
    e->default_timeout.tv_sec = delta/1000000;
    e->default_timeout.tv_usec = delta%1000000;
}
	
/*++++
 Add an IO descriptor to the internal list of `observed' descriptors.
 The given function will be called when the descriptor is ready for
 reading.

 Params:
	vp : event manager reference.
	chan : IO descriptor being inserted.
	f : function to be called.
	p : function arguments.
 Returns:
	0 on success, < 0 on failure, > 0 if a warning
	is generated, use rlc_error_info() to get error
	or warning description. Possible error/warning
	causes are:
		- TOO_MANY_IOS (too many IO descriptors,
		Check the compile time defined constant MAX_CHAN).
*/
int
rlc_insertchan_(void *vp, int chan, int id, int (*f)(void *), void *p)
{
#define	e ((evlist_t *)vp)
    int i, j, found;

    NDEB(fprintf(stderr, "rlc_insertchan: inserting chan %d\n",
	chan))
    /***
     *** check this...
     *** 17/10/97 lv fixed to handle holes...
     ***/
#if 1
    for (i=0, j=0, found = -1; j<MAX_CHAN && i<=e->io_ch_num; j++) {
	if (e->io_chan[j] < 0) {
	    /* get the first hole */
	    if (found < 0)
		found = j;
	} else {
	   i++;
	   if (e->io_chan[j]==chan) {
		/* found the same channel */
		found = j;
		break;
	   }
	}
    }
    if (found < 0) {
	return (rlc_internal_set_error_(-1, TOO_MANY_IOS, NULL));
    }
    i = found;
#else
    for (i=0; e->io_chan[i]!=chan && e->io_chan[i]>=0 && i<MAX_CHAN; i++);
#endif
    if (e->io_chan[i] != chan)
	(e->io_ch_num)++;
    e->io_chan[i]=chan;
    e->io_id[i]=id;
    e->io_f[i] = f;
    e->io_p[i] = p;
    return 0;
}

/*++++
 Removes an IO descriptor from the internal list.

 Params:
	vp : event manager reference.
	chan : IO descriptor.
*/
void
rlc_deletechan_(void *vp, int chan)
{
#define	e ((evlist_t *)vp)
    int i;

    NDEB(fprintf(stderr, "rlc_deletechan: deleting chan %d\n",
	chan))
    if (chan < 0) return;
    for (i=0; e->io_chan[i]!=chan && i<MAX_CHAN; i++);
    if (e->io_chan[i]==chan) {
	e->io_chan[i]=-1;
	(e->io_ch_num)--;
    }
    
}


/*++++
 Returns the IO descriptor last found ready by rlc_getevent.

 Params:
	vp : event manager reference.

 Returns:
	the IO descriptor last found ready.
*/
int
rlc_showselected_ (void *vp)
{
#define	e ((evlist_t *)vp)
    return(e->selected);
}

/*++++
 Arranges for a timer event to happen after delta usec.

 Params:
	vp : event manager reference.
	id : id of the event to be added, will be returned
		by rlc_getevent after the event execution. It must be
		$> 0$ to avoid confusion with predefined
		return values.
	delta : relative time of event execution (in usec).
	f : function to be called.
	p : function arguments.
	real : if set, the event will occurr delta usec
		after the actual calling time. Otherwise, it is
		assumed that the function has been called by the
		scheduler, and the event is scheduled exactly delta
		usec after the scheduling function had to be
		executed. This allows not to accumulate errors.

 Returns:
	0 on success, < 0 on failure. Possible error
	causes are: 
		- ALLOC_TIMIT_ERROR (Error allocating a timer
		descriptor)
*/
int
rlc_settimer_(void *vp, int id, unsigned long delta, CALLBACK f, void *p, int real)
{
#define	e ((evlist_t *)vp)
    timit_t *a, *pr = NULL, *t = e->timevent;
    struct timeval thistime;

    if ((a = new_timit()) == NULL) {
	return (rlc_internal_set_error_(-1, ALLOC_TIMIT_ERROR, NULL));
    }
    a->t_f = f;
    a->t_p = p;
    a->next = NULL;
    a->id = id;

    NDEB(fprintf(stderr, "<> scheduling event %d at %d\n", id, delta);)
    if (delta < e->late) {
	NDEB(fprintf(stderr, "ARGH we are scheduling event"
		" %d %u usec in the past!!!\n",
		id, e->late-delta));
    }

    if (real) {
	timestamp_t(&thistime);
	delta += diff_t( thistime, e->lasttime) ;
    }

    if (e->timevent == NULL) {
	a->delta = delta;
	e->timevent = a;
	return 0;
    }
	
    while(t!= NULL && delta > t->delta) {
	delta -= t->delta;
	pr = t;
	t = t->next;
    }

    a->delta = delta;
    a->next = t;
    if (t != NULL)
	(t->delta) -= delta;
    if (t == e->timevent) {
	e->timevent = a;
    } else {
	pr->next = a;
    }
    return 0;
}

/*++++
 Removes all the pending events with the given id,
 returns the number of events removed.

 Params:
	vp : event manager reference.
	id : id of event(s) to be removed.

 Returns:
	the number of events removed.
*/
int
rlc_deletetimer_(void *vp, int id)
{
#define	e ((evlist_t *)vp)
    timit_t *pr = NULL, *t = e->timevent;
    ui32 delay = 0;
    int how=0;

    /* if dummy event do nothing */
    if (id <= 0)
	return (0);

    while(t!= NULL) {
	if (t->id == id) {
	    timit_t *a = t->next;
	    if (pr==NULL)
		e->timevent = t->next;
	    else
		pr->next = t->next;
	    delay += t->delta;
	    free_timit(t);
	    /* XXX we still don't free argument (t->t_p) */
	    t = a;
	    how++;
	} else {
	    t->delta += delay;
	    pr = t;
	    t = t->next;
	}
    }
    return (how);
}


#ifdef TEST

int my_f(void *p)
{
   struct timeval time;
   timestamp_t(&time);

   printf("%d at %u\n", *(int *)p, time.tv_usec+(time.tv_sec%1000)*1000000);
   *(int *)p = *(int *)p +1;
   if (*(int *)p < 1000)
	rlc_settimer_(e, (*(int *)p)+1, 10000, my_f, p, 0);
}

/**
 ** arrange for a timer event to happen every delta msec
 **/
int main()
{
  int cont =0;
  void *e = rlc_createevlist_();

  rlc_settimer_(e, 1, 10000, my_f, (void *)&cont, 1);
  while (rlc_getevent_(e, 0)!=EV_NO_PENDING)
  	;
}

#endif /* TEST */
